from __future__ import annotations

import contextlib
import logging
import typing as t

import anyio
import httpx
from pyparsing import cast
from starlette.requests import Request

from _bentoml_sdk import Service
from bentoml import get_current_service
from bentoml._internal.utils import expand_envs
from bentoml.exceptions import BentoMLConfigException

if t.TYPE_CHECKING:
    from starlette.applications import Starlette

logger = logging.getLogger("bentoml.server")


async def _check_health(client: httpx.AsyncClient, health_endpoint: str) -> bool:
    try:
        response = await client.get(health_endpoint, timeout=5.0)
        if response.status_code == 404:
            raise BentoMLConfigException(
                f"Health endpoint {health_endpoint} not found (404). Please make sure the health "
                "endpoint is correctly configured in the service config."
            )
        return response.is_success
    except (httpx.HTTPError, httpx.RequestError):
        return False


def create_proxy_app(service: Service[t.Any]) -> Starlette:
    """A reverse-proxy that forwards all requests to the HTTP server started
    by the custom command.
    """
    import fastapi
    from fastapi.responses import StreamingResponse

    health_endpoint = service.config.get("endpoints", {}).get("livez", "/health")

    @contextlib.asynccontextmanager
    async def lifespan(
        app: fastapi.FastAPI,
    ) -> t.AsyncGenerator[dict[str, t.Any], None]:
        server_instance = get_current_service()
        assert server_instance is not None, "Current service is not initialized"
        async with contextlib.AsyncExitStack() as stack:
            if cmd_getter := getattr(server_instance, "__command__", None):
                if not callable(cmd_getter):
                    raise TypeError(
                        f"__command__ must be a callable that returns a list of strings, got {type(cmd_getter)}"
                    )
                cmd = cast("list[str]", cmd_getter())
            else:
                cmd = service.cmd
            assert cmd is not None, "must have a command"
            cmd = [expand_envs(c) for c in cmd]
            logger.info("Running service with command: %s", " ".join(cmd))
            if (
                instance_client := getattr(server_instance, "client", None)
            ) is not None and isinstance(instance_client, httpx.AsyncClient):
                # TODO: support aiohttp client
                client = instance_client
            else:
                proxy_port = service.config.get("http", {}).get("proxy_port", 8000)
                proxy_url = f"http://localhost:{proxy_port}"
                client = await stack.enter_async_context(
                    httpx.AsyncClient(base_url=proxy_url, timeout=None)
                )
            proc = await anyio.open_process(cmd, stdout=None, stderr=None)
            while proc.returncode is None:
                if await _check_health(client, health_endpoint):
                    break
                await anyio.sleep(0.5)
            else:
                raise RuntimeError(
                    "Service process exited before becoming healthy, see the error above"
                )

            app.state.client = client
            try:
                state = {"proc": proc, "client": client}
                service.context.state.update(state)
                yield state
            finally:
                proc.terminate()
                await proc.wait()

    assert service.has_custom_command(), "Service does not have custom command"
    app = fastapi.FastAPI(lifespan=lifespan)

    # TODO: support websocket endpoints
    @app.api_route(
        "/{path:path}",
        methods=["GET", "POST", "PUT", "DELETE", "PATCH", "OPTIONS", "HEAD"],
    )
    async def reverse_proxy(request: Request, path: str):
        url = httpx.URL(
            path=f"/{path}", query=request.url.query.encode("utf-8") or None
        )
        client = t.cast(httpx.AsyncClient, app.state.client)
        headers = dict(request.headers)
        headers.pop("host", None)
        req = client.build_request(
            method=request.method, url=url, headers=headers, content=request.stream()
        )
        try:
            resp = await client.send(req, stream=True)
        except httpx.ConnectError:
            return fastapi.Response(503)
        except httpx.RequestError:
            return fastapi.Response(500)

        return StreamingResponse(
            resp.aiter_raw(),
            status_code=resp.status_code,
            headers=resp.headers,
            background=resp.aclose,
        )

    return app



from __future__ import annotations

import inspect
import json
import logging
import signal
from typing import TYPE_CHECKING
from typing import Any
from typing import cast

import anyio
import click

if TYPE_CHECKING:
    from _bentoml_sdk import Service as _Service

    Service = _Service[Any]

logger = logging.getLogger("bentoml.worker.runner")


@click.command()
@click.argument("bento_identifier", type=click.STRING, required=False, default=".")
@click.option("--service-name", type=click.STRING, required=False, default="")
@click.option(
    "--worker-id",
    required=False,
    type=click.INT,
    default=None,
    help="If set, start the server as a bare worker with the given worker ID. Otherwise start a standalone server with a supervisor process.",
)
@click.option(
    "--development-mode",
    type=click.BOOL,
    help="Run the process in development mode",
    is_flag=True,
    default=False,
    show_default=True,
)
@click.option("--args", type=click.STRING, help="Bento arguments dict for the service")
def main(
    bento_identifier: str,
    service_name: str,
    worker_id: int | None,
    development_mode: bool,
    args: str | None,
):
    """Start a worker for running custom commands."""

    from _bentoml_impl.loader import load
    from bentoml._internal.container import BentoMLContainer
    from bentoml._internal.context import server_context
    from bentoml._internal.log import configure_server_logging
    from bentoml._internal.utils.args import set_arguments

    if args:
        set_arguments(json.loads(args))

    configure_server_logging()

    if worker_id is not None:
        server_context.worker_index = worker_id
    # standalone_build=True means to not restore cwd and model store
    service = load(bento_identifier, standalone_load=True)

    if service_name and service_name != service.name:
        service = service.find_dependent_by_name(service_name)
        server_context.service_type = "service"
    else:
        server_context.service_type = "entry_service"

    BentoMLContainer.development_mode.set(development_mode)

    server_context.service_name = service.name
    if service.bento is None:
        server_context.bento_name = service.name
        server_context.bento_version = "not available"
    else:
        server_context.bento_name = service.bento.tag.name
        server_context.bento_version = service.bento.tag.version or "not available"

    anyio.run(start_service, service)


async def start_service(service: Service) -> None:
    from _bentoml_sdk.service import set_current_service
    from bentoml._internal.utils import expand_envs

    instance = service.inner()

    if cmd_getter := getattr(instance, "__command__", None):
        if not callable(cmd_getter):
            raise TypeError(
                f"__command__ must be a callable that returns a list of strings, got {type(cmd_getter)}"
            )
        cmd = cast("list[str]", cmd_getter())
    else:
        cmd = service.cmd
    assert cmd is not None, "must have a command"
    cmd = [expand_envs(c) for c in cmd]
    logger.info("Running service with command: %s", " ".join(cmd))
    # Call on_startup hook with optional ctx or context parameter
    for name, member in vars(service.inner).items():
        if callable(member) and getattr(member, "__bentoml_startup_hook__", False):
            logger.info("Running startup hook: %s", name)
            result = getattr(instance, name)()  # call the bound method
            if inspect.isawaitable(result):
                await result
                logger.info("Completed async startup hook: %s", name)
            else:
                logger.info("Completed startup hook: %s", name)
    set_current_service(instance)
    process = None

    def signal_handler(sig, *_):
        if process is not None:
            process.send_signal(sig)

    original_handler = signal.signal(signal.SIGTERM, signal_handler)
    try:
        process = await anyio.open_process(cmd, stdout=None, stderr=None)
        retcode = await process.wait()
        logger.info("Process exited with code %d", retcode)
    finally:
        signal.signal(signal.SIGTERM, original_handler)
        # Call on_shutdown hook with optional ctx or context parameter
        for name, member in vars(service.inner).items():
            if callable(member) and getattr(member, "__bentoml_shutdown_hook__", False):
                logger.info("Running cleanup hook: %s", name)
                result = getattr(instance, name)()  # call the bound method
                if inspect.isawaitable(result):
                    await result
                    logger.info("Completed async cleanup hook: %s", name)
                else:
                    logger.info("Completed cleanup hook: %s", name)
        set_current_service(None)


if __name__ == "__main__":
    main()  # pylint: disable=no-value-for-parameter



from __future__ import annotations

import logging
import os
import posixpath
import shutil
import tempfile
import typing as t
import urllib.parse
import weakref
from abc import ABC
from abc import abstractmethod
from pathlib import Path

import fsspec
from fsspec.registry import known_implementations

from .types import PathType
from .utils.filesystem import safe_remove_dir

T = t.TypeVar("T", bound="Exportable")


class Exportable(ABC):
    _path: Path

    @staticmethod
    @abstractmethod
    def _export_ext() -> str:
        raise NotImplementedError

    @property
    @abstractmethod
    def _export_name(self) -> str:
        raise NotImplementedError

    @classmethod
    def _from_fs(cls, fs: fsspec.AbstractFileSystem) -> t.Self:
        """Create an instance from a file system."""
        temp_dir = tempfile.mkdtemp(prefix="bentoml-import-")
        fs.get("/", temp_dir, recursive=True)
        instance = cls.from_path(temp_dir)
        weakref.finalize(instance, safe_remove_dir, temp_dir)
        return instance

    @classmethod
    @abstractmethod
    def from_path(cls, path: PathType) -> t.Self:
        """Create an instance from a file path."""
        raise NotImplementedError

    @classmethod
    def guess_format(cls, path: str) -> str:
        _, ext = posixpath.splitext(path)

        if ext == "":
            return cls._export_ext()

        ext = ext[1:]
        if ext in [cls._export_ext(), "gz", "xz", "bz2", "tar", "zip"]:
            return ext
        else:
            return cls._export_ext()

    @classmethod
    def import_from(
        cls,
        path: str,
        input_format: str | None = None,
        *,
        protocol: str | None = None,
        user: str | None = None,
        passwd: str | None = None,
        params: dict[str, str] | None = None,
        subpath: str | None = None,
    ) -> t.Self:
        is_url = "://" in path
        if is_url and any(
            v is not None for v in [protocol, user, passwd, params, subpath]
        ):
            raise ValueError(
                "An FS URL was passed as the input path; all additional information should be passed as part of the URL."
            )
        parsedurl = urllib.parse.urlsplit(path)
        protocol = parsedurl.scheme if is_url else (protocol or "file")
        if protocol in ("osfs", "tar", "zip"):
            protocol = "file"
        if is_url:
            resource = parsedurl.netloc.rpartition("@")[2]
            subpath = parsedurl.path
            resource_url = urllib.parse.urlunsplit((protocol, *parsedurl[1:]))
        else:
            netloc = ""
            if user is not None:
                netloc += urllib.parse.quote(user)
            if passwd is not None:
                netloc += ":" + urllib.parse.quote(passwd)
            if netloc:
                netloc += "@"
            path = path.replace(os.sep, "/")
            subpath = subpath.replace(os.sep, "/") if subpath else subpath
            if subpath is None:
                if protocol == "file":
                    resource, subpath = "", path
                else:
                    resource, _, subpath = path.partition("/")
            else:
                resource = path
            netloc += resource
            url_tuple = (
                protocol,
                netloc,
                subpath,
                urllib.parse.urlencode(params or {}),
                "",
            )
            resource_url = urllib.parse.urlunsplit(url_tuple)

        if protocol not in known_implementations:
            raise ValueError(
                f"Unknown or unsupported protocol {protocol}. Some supported protocols are 'ftp', 's3', and 'file'."
            )

        if input_format is None:
            input_format = cls.guess_format(subpath)
        if protocol != "file":
            resource_url = f"filecache::{resource_url}"

        if input_format == "folder":
            from fsspec.implementations.dirfs import DirFileSystem

            fs, fspath = fsspec.url_to_fs(resource_url)
            if protocol == "file":  # Use the local file system
                return cls.from_path(fspath)
            fs = DirFileSystem(fspath, fs=fs)
        elif input_format == "zip":
            from fsspec.implementations.zip import ZipFileSystem

            fs = ZipFileSystem(resource_url)
        else:
            from fsspec.implementations.tar import TarFileSystem

            compressions = {
                cls._export_ext(): "xz",
                "gz": "gzip",
                "tar": None,
            }
            fs = TarFileSystem(
                resource_url, compression=compressions.get(input_format, input_format)
            )

        return cls._from_fs(fs)

    def export(
        self,
        path: str,
        output_format: t.Optional[str] = None,
        *,
        protocol: t.Optional[str] = None,
        user: t.Optional[str] = None,
        passwd: t.Optional[str] = None,
        params: t.Optional[t.Dict[str, str]] = None,
        subpath: t.Optional[str] = None,
    ) -> str:
        is_url = "://" in path
        if is_url and any(
            v is not None for v in [protocol, user, passwd, params, subpath]
        ):
            raise ValueError(
                "An FS URL was passed as the output path; all additional information should be passed as part of the URL."
            )
        parsedurl = urllib.parse.urlsplit(path)
        protocol = parsedurl.scheme if is_url else (protocol or "file")
        if protocol in ("osfs", "tar", "zip"):
            protocol = "file"
        if is_url:
            resource = parsedurl.netloc.rpartition("@")[2]
            subpath = parsedurl.path
            netloc = parsedurl.netloc
            query = parsedurl.query
        else:
            netloc = ""
            if user is not None:
                netloc += urllib.parse.quote(user)
            if passwd is not None:
                netloc += ":" + urllib.parse.quote(passwd)
            if netloc:
                netloc += "@"
            path = path.replace(os.sep, "/")
            subpath = subpath.replace(os.sep, "/") if subpath else subpath
            if subpath is None:
                if protocol == "file":
                    resource, subpath = "", path
                else:
                    resource, _, subpath = path.partition("/")
            else:
                resource = path
            netloc += resource
            query = urllib.parse.urlencode(params or {})

        if protocol not in known_implementations:
            raise ValueError(
                f"Unknown or unsupported protocol {protocol}. Some supported protocols are 'ftp', 's3', and 'file'."
            )

        if output_format is None:
            output_format = self.guess_format(subpath)

        is_dir = (
            not subpath
            or subpath.endswith("/")
            or protocol == "file"
            and os.path.isdir(subpath)
        )

        if output_format != "folder" and is_dir:
            subpath = posixpath.join(
                subpath or "", f"{self._export_name}.{output_format}"
            )
        if output_format == "folder" and not subpath.endswith("/"):
            subpath += "/"

        if output_format == self._export_ext() and not subpath.endswith(
            f".{self._export_ext()}"
        ):
            logging.info(
                f"Adding {self._export_ext()} because ext is {posixpath.splitext(subpath)[1]}"
            )
            subpath += f".{self._export_ext()}"
        resource_url = urllib.parse.urlunsplit((protocol, netloc, subpath, query, ""))

        fs, fspath = fsspec.url_to_fs(resource_url)
        if output_format == "folder":
            fs.put(str(self._path), fspath, recursive=True)
        else:
            temp_name = tempfile.mktemp(prefix="bentoml-export-")
            compressed = self._compress(temp_name, output_format)
            try:
                fs.put(compressed, fspath)
            finally:
                os.remove(compressed)
        return fspath

    def _compress(self, path: str, output_format: str) -> str:
        formats = {
            "gz": "gztar",
            "xz": "xztar",
            "bz2": "bztar",
            self._export_ext(): "xztar",
        }
        if output_format in ["gz", "xz", "bz2", "tar", self._export_ext(), "zip"]:
            return shutil.make_archive(
                path, formats.get(output_format, output_format), str(self._path)
            )
        else:
            raise ValueError(f"Unsupported format {output_format}")



import contextlib
import ipaddress
import os
import pathlib
import socket
from typing import no_type_check
from urllib.parse import quote
from urllib.parse import unquote
from urllib.parse import urlparse
from urllib.request import url2pathname

import psutil


def path_to_uri(path: str) -> str:
    """
    Convert a path to a URI.

    Args:
        path: Path to convert to URI.

    Returns:
        URI string. (quoted, absolute)
    """
    path = os.path.abspath(path)
    if psutil.WINDOWS:
        return pathlib.PureWindowsPath(path).as_uri()
    if psutil.POSIX:
        return pathlib.PurePosixPath(path).as_uri()
    raise ValueError("Unsupported OS")


def uri_to_path(uri: str) -> str:
    """
    Convert a file URI to a path.

    Args:
        uri: URI to convert to path.

    Returns:
        Path string. (unquoted)
    """
    parsed = urlparse(uri)
    if parsed.scheme not in ("file", "filesystem", "unix"):
        raise ValueError("Unsupported URI scheme")
    host = "{0}{0}{mnt}{0}".format(os.path.sep, mnt=parsed.netloc)
    return os.path.normpath(os.path.join(host, url2pathname(unquote(parsed.path))))


def encode_path_for_uri(path: str) -> str:
    """Percent-encode non-URL characters in a path."""
    return quote(path.replace(os.sep, "/"))


def is_http_url(url: str) -> bool:
    return urlparse(url).scheme in {"http", "https"}


original_create_connection = None


@contextlib.contextmanager
def make_safe_connect():
    """Patch loop.create_connection() method to reject unsafe URLs."""

    from urllib.request import getproxies

    import httpx
    from uvloop import Loop

    from bentoml.exceptions import BadInput

    global original_create_connection

    if original_create_connection is None:
        original_create_connection = Loop.create_connection

    # Do not check connections with proxy servers
    proxies = [
        (parsed.hostname, parsed.port)
        for parsed in map(urlparse, getproxies().values())
    ]

    @no_type_check
    async def safe_create_connection(
        self, protocol_factory, host=None, port=None, **kwargs
    ):
        if host is not None and (host, port) not in proxies:
            try:
                ip = ipaddress.ip_address(host)
            except ValueError:
                raise socket.gaierror(f"Blocked invalid IP address {host}")
            else:
                if ip.is_private or ip.is_loopback or ip.is_link_local:
                    raise socket.gaierror(f"Blocked private IP address {host}")
        return await original_create_connection(
            self, protocol_factory, host=host, port=port, **kwargs
        )

    Loop.create_connection = safe_create_connection
    try:
        yield
    except httpx.ConnectError as e:
        if "All connection attempts failed" in str(e):
            raise BadInput("Connection blocked due to insecure input URL") from e
    finally:
        Loop.create_connection = original_create_connection



from __future__ import annotations

import datetime
import shutil
import typing as t
from abc import abstractmethod
from contextlib import contextmanager
from pathlib import Path

from ..exceptions import BentoMLException
from ..exceptions import NotFound
from .exportable import Exportable
from .tag import Tag
from .types import PathType
from .utils.filesystem import calc_dir_size

T = t.TypeVar("T")


class StoreItem(Exportable):
    @property
    @abstractmethod
    def tag(self) -> Tag:
        raise NotImplementedError

    @classmethod
    def get_typename(cls) -> str:
        return cls.__name__

    @property
    def _export_name(self) -> str:
        return f"{self.tag.name}-{self.tag.version}"

    @property
    @abstractmethod
    def creation_time(self) -> datetime.datetime:
        raise NotImplementedError

    @property
    def path(self) -> str:
        return str(self._path)

    def path_of(self, item: str) -> str:
        return str(self._path / item.lstrip("/"))

    @property
    def file_size(self) -> int:
        return calc_dir_size(self._path)

    def __repr__(self):
        return f'{self.get_typename()}(tag="{self.tag}")'


Item = t.TypeVar("Item", bound=StoreItem)


class Store(t.Generic[Item]):
    """A Store manages items under the given base filesystem.

    Note that Store has no consistency checks; it assumes that no direct modification
    of the files in its directory has occurred.

    """

    _item_type: t.Type[Item]

    def __init__(self, base_path: PathType):
        self._path = Path(base_path)

    def list(self, tag: Tag | str | None = None) -> t.List[Item]:
        if not tag:
            # Return all items in the store
            return [
                ver
                for _d in sorted(self._path.iterdir())
                if _d.is_dir()
                for ver in self.list(_d.name)
            ]

        _tag = Tag.from_taglike(tag)
        if _tag.version is None:
            item_dir = self._path / _tag.name
            if not item_dir.is_dir():
                raise NotFound(
                    f"no {self._item_type.get_typename()}s with name '{_tag.name}' found"
                )

            tags = sorted(
                Tag(_tag.name, f.name) for f in item_dir.iterdir() if f.is_dir()
            )
            return [self._get_item(t) for t in tags]
        else:
            return [self._get_item(_tag)] if (self._path / _tag.path()).is_dir() else []

    def _get_item(self, tag: Tag) -> Item:
        """
        Creates a new instance of Item that represents the item with tag `tag`.
        """
        return self._item_type.from_path(self._path / tag.path())

    def _recreate_latest(self, tag: Tag):
        try:
            items = self.list(tag.name)
        except NotFound:
            raise NotFound(
                f"no {self._item_type.get_typename()}s with name '{tag.name}' exist in BentoML store {self._path}"
            )

        if len(items) == 0:
            raise NotFound(
                f"no {self._item_type.get_typename()}s with name '{tag.name}' exist in BentoML store {self._path}"
            )

        items.sort(reverse=True, key=lambda item: item.creation_time)
        latest_version = t.cast(str, items[0].tag.version)

        with (self._path / tag.latest_path()).open("w") as latest_file:
            latest_file.write(latest_version)
        tag.version = latest_version

    def get(self, tag: Tag | str) -> Item:
        """
        store.get("my_bento")
        store.get("my_bento:v1.0.0")
        store.get(Tag("my_bento", "latest"))
        """
        _tag = Tag.from_taglike(tag)
        if _tag.version is None or _tag.version == "latest":
            try:
                _tag.version = (self._path / _tag.latest_path()).read_text().strip()

                if not (self._path / _tag.path()).exists():
                    self._recreate_latest(_tag)
            except OSError:
                self._recreate_latest(_tag)

        path = _tag.path()
        if (self._path / path).exists():
            return self._get_item(_tag)

        matches = list(p for p in self._path.glob(f"{path}*/") if p.is_dir())
        if len(matches) == 1:
            return self._get_item(Tag(_tag.name, matches[0].name))
        elif len(matches) > 1:
            vers = [p.name for p in matches]
            raise BentoMLException(
                f"Multiple {_tag.name} versions found: {', '.join(vers)}"
            )
        # No match found
        cmd = (
            "bentoml models pull"
            if self._item_type.get_typename() == "Model"
            else "bentoml pull"
        )
        raise NotFound(
            f"{self._item_type.get_typename()} '{tag}' is not found in BentoML store {self._path}, "
            f"you may need to run `{cmd}` first"
        )

    @contextmanager
    def register(self, tag: str | Tag) -> t.Generator[str, None, None]:
        _tag = Tag.from_taglike(tag)

        item_path = _tag.path()
        if (self._path / item_path).exists():
            raise BentoMLException(
                f"Item '{_tag}' already exists in the store {self._path}"
            )
        (self._path / item_path).mkdir(parents=True, exist_ok=True)
        try:
            yield str(self._path / item_path)
        except Exception:
            shutil.rmtree(self._path / item_path)
            raise
        # item generation is most likely successful, link latest path
        if (
            not (latest_file := self._path / _tag.latest_path()).exists()
            or self.get(_tag).creation_time >= self.get(_tag.name).creation_time
        ):
            with latest_file.open("w") as f:
                f.write(_tag.version)

    def delete(self, tag: str | Tag) -> None:
        _tag = Tag.from_taglike(tag)

        if _tag.version == "latest":
            try:
                _tag.version = (self._path / _tag.latest_path()).read_text().strip()
            except OSError:
                # if latest path doesn't exist, we don't need to delete anything
                return

        tag_path = self._path / _tag.path()
        if not tag_path.exists():
            raise NotFound(f"{self._item_type.get_typename()} '{tag}' not found")

        shutil.rmtree(tag_path)
        if (self._path / _tag.name).is_dir():
            versions = self.list(_tag.name)
            if len(versions) == 0:
                # if we've removed all versions, remove the directory
                shutil.rmtree(self._path / _tag.name)
            else:
                new_latest = sorted(versions, key=lambda x: x.creation_time)[-1]
                # otherwise, update the latest version
                assert new_latest.tag.version is not None
                (self._path / _tag.latest_path()).write_text(new_latest.tag.version)



from __future__ import annotations

import logging
import os
import subprocess
import typing as t
from abc import ABC
from abc import abstractmethod
from pathlib import Path
from shutil import which
from tempfile import NamedTemporaryFile

import attr

from ...exceptions import BentoMLException
from ..bento.bento import Bento
from ..configuration import get_debug_mode

logger = logging.getLogger(__name__)


@attr.define
class Environment(ABC):
    name: str
    path: Path = attr.field(converter=Path)
    # path to bento's /env dir
    bento: Bento
    env_exe: str = attr.field(init=False)

    def __attrs_post_init__(self):
        self.env_exe = self.get_executable()
        if not self.path.exists():
            self.create()

    @abstractmethod
    def get_executable(self) -> str:
        """
        Returns executable path responsible for running this environment.

        Make sure that said executable is available in PATH.
        """
        ...

    @abstractmethod
    def create(self):
        """
        Create the environment with the files from bento.
        """
        ...

    @abstractmethod
    def run(self, commands: list[str]):
        """
        run the commands in an activated environment.
        """
        ...

    @staticmethod
    def run_script_subprocess(
        script_file_path: t.Union[str, os.PathLike[str]],
        capture_output: bool,
        debug_mode: bool,
    ):
        shell_path = which("bash")
        if shell_path is None:
            raise BentoMLException("Unable to locate a valid shell")

        safer_bash_args: list[str] = []

        # but only work in bash
        if debug_mode:
            safer_bash_args = ["-euxo", "pipefail"]
        result = subprocess.run(
            [shell_path, *safer_bash_args, script_file_path],
            capture_output=capture_output,
        )
        if result.returncode != 0:
            if result.stdout:
                logger.debug(result.stdout.decode())
            if result.stderr:
                logger.error(result.stderr.decode())
            raise BentoMLException(
                "Subprocess call returned non-zero value. Reffer logs for more details"
            )


class Conda(Environment):
    def get_executable(self) -> str:
        conda_exe = os.environ.get("CONDA_EXE")
        if conda_exe is None:
            raise BentoMLException(
                "Conda|Miniconda executable not found! Make sure any one is installed and environment is activated."
            )
        return conda_exe

    def create(self):
        # install conda deps
        from ..bento.build_config import CONDA_ENV_YAML_FILE_NAME

        # create a env under $BENTOML_HOME/env
        # setup conda with bento's environment.yml file and python/install.sh file
        conda_env_path = self.path / self.name
        python_version: str
        with open(self.bento.path_of("/env/python/version.txt"), "r") as pyver_file:
            py_version = pyver_file.read().split(".")[:2]
            python_version = ".".join(py_version)
        conda_environment_file = self.bento.path_of(
            f"/env/conda/{CONDA_ENV_YAML_FILE_NAME}"
        )
        create_script = f"""\
eval "$({self.get_executable()} shell.posix hook)"

conda create -p {conda_env_path} python={python_version} --yes

if [ -f {conda_environment_file} ]; then
    conda config --set pip_interop_enabled True
    conda env update -p {conda_env_path} --file {conda_environment_file}
fi

conda activate {conda_env_path}
bash -euxo pipefail {self.bento.path_of("/env/python/install.sh")}
"""
        with NamedTemporaryFile(mode="w", delete=False) as script_file:
            script_file.write(create_script)

        logger.info("Creating Conda env and installing dependencies...")
        self.run_script_subprocess(
            script_file.name,
            capture_output=not get_debug_mode(),
            debug_mode=get_debug_mode(),
        )

    def run(self, commands: list[str]):
        """
        Run commands in the activated environment.
        """
        conda_env_path = self.path / self.name
        create_script = f"""\
eval "$({self.env_exe} shell.posix hook)"
conda activate {conda_env_path}
{" ".join(commands)}
"""
        with NamedTemporaryFile(mode="w", delete=False) as script_file:
            script_file.write(create_script)
        self.run_script_subprocess(
            script_file.name, capture_output=False, debug_mode=get_debug_mode()
        )
